from openmc_plasma_source import FusionRingSource

from openmc import Source
import pytest


def test_creation():
    my_source = FusionRingSource(radius=1, z_placement=1)
    assert isinstance(my_source, Source)


def test_wrong_fuel():
    with pytest.raises(ValueError):
        FusionRingSource(radius=1, fuel="топливо")



import openmc
import numpy as np


class FusionRingSource(openmc.Source):
    """An openmc.Source object with some presets to make it more convenient
    for fusion simulations using a ring source. All attributes can be changed
    after initialization if required. Default isotropic ring source with a Muir
    energy distribution.

    Args:
        radius: the inner radius of the ring source
        angles (iterable of floats): the start and stop angles of the ring in radians,
        z_placement: Location of the ring source (m). Defaults to 0.
        temperature: the temperature to use in the Muir distribution in eV,
    """

    def __init__(
        self,
        radius,
        angles=(0, 2 * np.pi),
        z_placement=0,
        temperature: float = 20000.0,
        fuel="DT",
    ):

        super().__init__()

        # performed after the super init as these are Source attributes
        radius = openmc.stats.Discrete([radius], [1])
        z_values = openmc.stats.Discrete([z_placement], [1])
        angle = openmc.stats.Uniform(a=angles[0], b=angles[1])
        self.space = openmc.stats.CylindricalIndependent(
            r=radius, phi=angle, z=z_values, origin=(0.0, 0.0, 0.0)
        )
        self.angle = openmc.stats.Isotropic()
        if fuel == "DT":
            mean_energy = 14080000.0  # mean energy in eV
            mass_of_reactants = 5  # mass of the reactants (D + T) AMU
        elif fuel == "DD":
            mean_energy = 2450000.0  # mean energy in eV
            mass_of_reactants = 4  # mass of the reactants (D + D) AMU
        else:
            raise ValueError(f'fuel must be either "DT" or "DD", not {fuel}')
        self.energy = openmc.stats.Muir(
            e0=mean_energy, m_rat=mass_of_reactants, kt=temperature
        )



import matplotlib.pyplot as plt
from matplotlib import cm
import numpy as np


def scatter_tokamak_source(source, quantity=None, **kwargs):
    """Create a 2D scatter plot of the tokamak source.
    See https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.scatter.html
    for more arguments.

    Args:
        source (ops.TokamakSource): the plasma source
        quantity ("str", optional): value by which the lines should be
            coloured. Defaults to None.

    Raises:
        ValueError: If the quantity is unknown
    """

    quantity_to_attribute = {
        "ion_temperature": source.temperatures,
        "neutron_source_density": source.neutron_source_density,
        "strength": source.strengths,
    }
    if quantity in quantity_to_attribute:
        colours = quantity_to_attribute[quantity]
    elif quantity is None:
        colours = None
    else:
        raise ValueError("Unknown quantity")
    plt.gca().set_aspect("equal")

    return plt.scatter(source.RZ[0], source.RZ[1], c=colours, **kwargs)


def plot_tokamak_source_3D(
    source, quantity=None, angles=[0, 1 / 2 * np.pi], colorbar="viridis", **kwargs
):
    """Creates a 3D plot of the tokamak source.
    See https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.plot.html#matplotlib.pyplot.plot
    for more arguments.

    Args:
        source (ops.TokamakSource): the plasma source
        quantity ("str", optional): value by which the lines should be
            coloured. Defaults to None.
        angles (list, optional): iterable of two floats defining the coverage.
            Defaults to [0, 1/2*np.pi].
        colorbar (str, optional): colorbar used if quantity is not None.
            Defaults to "viridis".

    Raises:
        ValueError: If the quantity is unknown
    """

    quantity_to_attribute = {
        "ion_temperature": source.temperatures,
        "neutron_source_density": source.neutron_source_density,
        "strength": source.strengths,
    }
    if quantity in quantity_to_attribute:
        values = quantity_to_attribute[quantity]
    elif quantity is None:
        values = None
    else:
        raise ValueError("Unknown quantity")

    colorbar = cm.get_cmap(colorbar)
    axes = plt.axes(projection="3d")
    theta = np.linspace(*angles, 100)
    for i in range(source.sample_size):
        if values is not None:
            colour = colorbar(values[i] / max(values))
        else:
            colour = None
        x = source.RZ[0][i] * np.sin(theta)
        y = source.RZ[0][i] * np.cos(theta)
        z = source.RZ[1][i]
        plt.plot(x, y, z, color=colour, **kwargs)

    axes.set_xlim(-source.major_radius, source.major_radius)
    axes.set_ylim(-source.major_radius, source.major_radius)
    axes.set_zlim(-source.major_radius, source.major_radius)



try:
    from importlib.metadata import version, PackageNotFoundError
except (ModuleNotFoundError, ImportError):
    from importlib_metadata import version, PackageNotFoundError
try:
    __version__ = version("openmc_plasma_source")
except PackageNotFoundError:
    from setuptools_scm import get_version

    __version__ = get_version(root="..", relative_to=__file__)

__all__ = ["__version__"]

from .tokamak_source import TokamakSource
from .ring_source import FusionRingSource
from .point_source import FusionPointSource



import openmc
from openmc_plasma_source import FusionPointSource

my_source = FusionPointSource(coordinate=(0, 0, 0), temperature=20000.0, fuel="DT")

# Create a single material
iron = openmc.Material()
iron.set_density("g/cm3", 5.0)
iron.add_element("Fe", 1.0)
mats = openmc.Materials([iron])

# Create a 5 cm x 5 cm box filled with iron
cells = []
inner_box1 = openmc.ZCylinder(r=600.0)
inner_box2 = openmc.ZCylinder(r=1400.0)
outer_box = openmc.model.rectangular_prism(4000.0, 4000.0, boundary_type="vacuum")
cells += [openmc.Cell(fill=iron, region=-inner_box1)]
cells += [openmc.Cell(fill=None, region=+inner_box1 & -inner_box2)]
cells += [openmc.Cell(fill=iron, region=+inner_box2 & outer_box)]
geometry = openmc.Geometry(cells)

# Tell OpenMC we're going to use our custom source
settings = openmc.Settings()
settings.run_mode = "fixed source"
settings.batches = 10
settings.particles = 1000
settings.source = my_source

# Finally, define a mesh tally so that we can see the resulting flux
mesh = openmc.RegularMesh()
mesh.lower_left = (-2000.0, -2000.0)
mesh.upper_right = (2000.0, 2000.0)
mesh.dimension = (50, 50)

tally = openmc.Tally()
tally.filters = [openmc.MeshFilter(mesh)]
tally.scores = ["flux"]
tallies = openmc.Tallies([tally])

model = openmc.model.Model(
    materials=mats, geometry=geometry, settings=settings, tallies=tallies
)

model.run()



import openmc
from openmc_plasma_source import TokamakSource


# create a plasma source
my_plasma = TokamakSource(
    elongation=1.557,
    ion_density_centre=1.09e20,
    ion_density_peaking_factor=1,
    ion_density_pedestal=1.09e20,
    ion_density_separatrix=3e19,
    ion_temperature_centre=45.9,
    ion_temperature_peaking_factor=8.06,
    ion_temperature_pedestal=6.09,
    ion_temperature_separatrix=0.1,
    major_radius=9.06,
    minor_radius=2.92258,
    pedestal_radius=0.8 * 2.92258,
    mode="H",
    shafranov_factor=0.44789,
    triangularity=0.270,
    ion_temperature_beta=6,
)


# Create a single material
iron = openmc.Material()
iron.set_density("g/cm3", 5.0)
iron.add_element("Fe", 1.0)
mats = openmc.Materials([iron])

# Create a 5 cm x 5 cm box filled with iron
cells = []
inner_box1 = openmc.ZCylinder(r=600.0)
inner_box2 = openmc.ZCylinder(r=1400.0)
outer_box = openmc.model.rectangular_prism(4000.0, 4000.0, boundary_type="vacuum")
cells += [openmc.Cell(fill=iron, region=-inner_box1)]
cells += [openmc.Cell(fill=None, region=+inner_box1 & -inner_box2)]
cells += [openmc.Cell(fill=iron, region=+inner_box2 & outer_box)]
geometry = openmc.Geometry(cells)

# Tell OpenMC we're going to use our custom source
settings = openmc.Settings()
settings.run_mode = "fixed source"
settings.batches = 10
settings.particles = 1000
settings.source = my_plasma.sources

# Finally, define a mesh tally so that we can see the resulting flux
mesh = openmc.RegularMesh()
mesh.lower_left = (-2000.0, -2000.0)
mesh.upper_right = (2000.0, 2000.0)
mesh.dimension = (50, 50)

tally = openmc.Tally()
tally.filters = [openmc.MeshFilter(mesh)]
tally.scores = ["flux"]
tallies = openmc.Tallies([tally])

model = openmc.model.Model(
    materials=mats, geometry=geometry, settings=settings, tallies=tallies
)

model.run()



import matplotlib.pyplot as plt
from openmc_plasma_source import plotting, TokamakSource
import pytest

my_plasma = TokamakSource(
    elongation=1.557,
    ion_density_centre=1.09e20,
    ion_density_peaking_factor=1,
    ion_density_pedestal=1.09e20,
    ion_density_separatrix=3e19,
    ion_temperature_centre=45.9,
    ion_temperature_peaking_factor=8.06,
    ion_temperature_pedestal=6.09,
    ion_temperature_separatrix=0.1,
    major_radius=9.06,
    minor_radius=2.92258,
    pedestal_radius=0.8 * 2.92258,
    mode="H",
    shafranov_factor=0.44789,
    triangularity=0.270,
    ion_temperature_beta=6,
)
my_plasma.sample_sources()


def test_scatter_tokamak_source():
    plt.figure()
    plotting.scatter_tokamak_source(my_plasma)
    plotting.scatter_tokamak_source(my_plasma, quantity="ion_temperature")
    plotting.scatter_tokamak_source(my_plasma, alpha=0.2)


def test_scatter_tokamak_wrong_quantity():
    plt.figure()
    with pytest.raises(ValueError):
        plotting.scatter_tokamak_source(my_plasma, quantity="coucou")


def test_plot_tokamak_source():
    plt.figure()
    plotting.plot_tokamak_source_3D(my_plasma)
    plotting.plot_tokamak_source_3D(my_plasma, quantity="ion_temperature")
    plotting.plot_tokamak_source_3D(my_plasma, alpha=0.2)
    plotting.plot_tokamak_source_3D(my_plasma, angles=[0, 3.14])


def test_plot_tokamak_wrong_quantity():
    plt.figure()
    with pytest.raises(ValueError):
        plotting.plot_tokamak_source_3D(my_plasma, quantity="coucou")



from typing import Tuple

import openmc


class FusionPointSource(openmc.Source):
    """An openmc.Source object with some presets to make it more convenient
    for fusion simulations using a point source. All attributes can be changed
    after initialization if required. Default isotropic point source at the
    origin with a Muir energy distribution.
    """

    def __init__(
        self,
        coordinate: Tuple[float, float, float] = (0, 0, 0),
        temperature: float = 20000.0,
        fuel: str = "DT",
    ):

        super().__init__()

        # performed after the super init as these are Source attributes
        self.space = openmc.stats.Point(coordinate)
        self.angle = openmc.stats.Isotropic()
        if fuel == "DT":
            mean_energy = 14080000.0  # mean energy in eV
            mass_of_reactants = 5  # mass of the reactants (D + T) AMU
        elif fuel == "DD":
            mean_energy = 2450000.0  # mean energy in eV
            mass_of_reactants = 4  # mass of the reactants (D + D) AMU
        else:
            raise ValueError(f'fuel must be either "DT" or "DD", not {fuel}')
        self.energy = openmc.stats.Muir(
            e0=mean_energy, m_rat=mass_of_reactants, kt=temperature
        )



import numpy as np


class TokamakSource:
    """Plasma neutron source sampling.
    This class greatly relies on models described in [1]

    [1] : Fausser et al, 'Tokamak D-T neutron source models for different
    plasma physics confinement modes', Fus. Eng. and Design,
    https://doi.org/10.1016/j.fusengdes.2012.02.025

    Usage:
        my_plasma = Plasma(**plasma_prms)
        my_plasma.sample_sources()
        print(my_plasma.RZ)
        print(my_plasma.temperatures)
        openmc_sources = my_plasma.make_openmc_sources()

    Args:
        major_radius (float): Plasma major radius (m)
        minor_radius (float): Plasma minor radius (m)
        elongation (float): Plasma elongation
        triangularity (float): Plasma triangularity
        mode (str): Confinement mode ("L", "H", "A")
        ion_density_centre (float): Ion density at the plasma centre (m-3)
        ion_density_peaking_factor (float): Ion density peaking factor
            (referred in [1] as ion density exponent)
        ion_density_pedestal (float): Ion density at pedestal (m-3)
        ion_density_separatrix (float): Ion density at separatrix (m-3)
        ion_temperature_centre (float): Ion temperature at the plasma
            centre (keV)
        ion_temperature_peaking_factor (float): Ion temperature peaking
            factor (referred in [1] as ion temperature exponent alpha_T)
        ion_temperature_beta (float): Ion temperature beta exponent
            (referred in [1] as ion temperature exponent beta_T)
        ion_temperature_pedestal (float): Ion temperature at pedestal (keV)
        ion_temperature_separatrix (float): Ion temperature at separatrix
            (keV)
        pedestal_radius (float): Minor radius at pedestal (m)
        shafranov_factor (float): Shafranov factor (referred in [1] as esh)
            also known as outward radial displacement of magnetic surfaces
            (m)
        angles (iterable of floats): the start and stop angles of the ring in
            radians
        sample_size (int, optional): number of neutron sources. Defaults
            to 1000.
    """

    def __init__(
        self,
        major_radius,
        minor_radius,
        elongation,
        triangularity,
        mode,
        ion_density_centre,
        ion_density_peaking_factor,
        ion_density_pedestal,
        ion_density_separatrix,
        ion_temperature_centre,
        ion_temperature_peaking_factor,
        ion_temperature_beta,
        ion_temperature_pedestal,
        ion_temperature_separatrix,
        pedestal_radius,
        shafranov_factor,
        angles=(0, 2 * np.pi),
        sample_size=1000,
    ) -> None:
        self.major_radius = major_radius
        self.minor_radius = minor_radius
        self.elongation = elongation
        self.triangularity = triangularity
        self.ion_density_centre = ion_density_centre
        self.mode = mode
        self.ion_density_peaking_factor = ion_density_peaking_factor
        self.pedestal_radius = pedestal_radius
        self.ion_density_pedestal = ion_density_pedestal
        self.ion_density_separatrix = ion_density_separatrix

        self.ion_temperature_centre = ion_temperature_centre
        self.ion_temperature_peaking_factor = ion_temperature_peaking_factor
        self.ion_temperature_pedestal = ion_temperature_pedestal
        self.ion_temperature_separatrix = ion_temperature_separatrix
        self.ion_temperature_beta = ion_temperature_beta

        self.shafranov_factor = shafranov_factor

        self.sample_size = sample_size

        self.angles = angles

        self.sample_sources()
        self.sources = self.make_openmc_sources()

    def ion_density(self, r):
        """Computes the ion density at a given position. The ion density is
        only dependent on the minor radius.

        Args:
            r (float, ndarray): the minor radius (m)

        Returns:
            float, ndarray: ion density in m-3
        """
        if self.mode == "L":
            density = (
                self.ion_density_centre
                * (1 - (r / self.major_radius) ** 2) ** self.ion_density_peaking_factor
            )
        elif self.mode in ["H", "A"]:
            # TODO: find an alternative to iterating through the array
            if isinstance(r, np.ndarray):
                density = []
                for radius in r:
                    if 0 < radius < self.pedestal_radius:
                        prod = self.ion_density_centre - self.ion_density_pedestal
                        prod *= (
                            1 - (radius / self.pedestal_radius) ** 2
                        ) ** self.ion_density_peaking_factor

                        density_loc = self.ion_density_pedestal + prod
                    else:
                        prod = self.ion_density_pedestal - self.ion_density_separatrix
                        prod *= (self.major_radius - radius) / (
                            self.major_radius - self.pedestal_radius
                        )
                        density_loc = self.ion_density_separatrix + prod
                    density.append(density_loc)
                density = np.array(density)
        return density

    def ion_temperature(self, r):
        """Computes the ion temperature at a given position. The ion
        temperature is only dependent on the minor radius.

        Args:
            r (float, ndarray): minor radius (m)

        Returns:
            float, ndarray: ion temperature (keV)
        """
        if self.mode == "L":
            temperature = (
                self.ion_temperature_centre
                * (1 - (r / self.major_radius) ** 2)
                ** self.ion_temperature_peaking_factor
            )
        elif self.mode in ["H", "A"]:
            # TODO: find an alternative to iterating through the array
            if isinstance(r, np.ndarray):
                temperature = []
                for radius in r:
                    if 0 < radius < self.pedestal_radius:
                        prod = (
                            self.ion_temperature_centre - self.ion_temperature_pedestal
                        )
                        prod *= (
                            1
                            - (radius / self.pedestal_radius)
                            ** self.ion_temperature_beta
                        ) ** self.ion_temperature_peaking_factor

                        temperature_loc = self.ion_temperature_pedestal + prod
                    else:
                        prod = (
                            self.ion_temperature_pedestal
                            - self.ion_temperature_separatrix
                        )
                        prod *= (self.major_radius - radius) / (
                            self.major_radius - self.pedestal_radius
                        )
                        temperature_loc = self.ion_temperature_separatrix + prod
                    temperature.append(temperature_loc)
                temperature = np.array(temperature)
        return temperature

    def convert_a_alpha_to_R_Z(self, a, alpha):
        """Converts (r, alpha) cylindrical coordinates to (R, Z) cartesian
        coordinates.

        Args:
            a (float, ndarray): minor radius (m)
            alpha (float, ndarray): angle (rad)

        Returns:
            ((float, ndarray), (float, ndarray)): (R, Z) coordinates
        """
        shafranov_shift = self.shafranov_factor * (1.0 - (a / self.minor_radius) ** 2)
        R = (
            self.major_radius
            + a * np.cos(alpha + (self.triangularity * np.sin(alpha)))
            + shafranov_shift
        )
        Z = self.elongation * a * np.sin(alpha)
        return (R, Z)

    def sample_sources(self):
        """Samples self.sample_size neutrons and creates attributes .densities
        (ion density), .temperatures (ion temperature), .strengths
        (neutron source density) and .RZ (coordinates)
        """
        # create a sample of (a, alpha) coordinates
        a = np.random.random(self.sample_size) * self.minor_radius
        alpha = np.random.random(self.sample_size) * 2 * np.pi

        # compute densities, temperatures, neutron source densities and
        # convert coordinates
        self.densities = self.ion_density(a)
        self.temperatures = self.ion_temperature(a)
        self.neutron_source_density = neutron_source_density(
            self.densities, self.temperatures
        )
        self.strengths = self.neutron_source_density / sum(self.neutron_source_density)
        self.RZ = self.convert_a_alpha_to_R_Z(a, alpha)

    def make_openmc_sources(self):
        """Creates a list of OpenMC Sources() objects. The created sources are
        ring sources based on the .RZ coordinates between two angles. The
        energy of the sources are Muir energy spectra with ion temperatures
        based on .temperatures. The strength of the sources (their probability)
        is based on .strengths.

        Args:
            angles ((float, float), optional): rotation of the ring source.
            Defaults to (0, 2*np.pi).

        Returns:
            list: list of openmc.Source()
        """
        import openmc

        sources = []
        # create a ring source for each sample in the plasma source
        for i in range(len(self.strengths)):
            my_source = openmc.Source()

            # extract the RZ values accordingly
            radius = openmc.stats.Discrete([self.RZ[0][i]], [1])
            z_values = openmc.stats.Discrete([self.RZ[1][i]], [1])
            angle = openmc.stats.Uniform(a=self.angles[0], b=self.angles[1])

            # create a ring source
            my_source.space = openmc.stats.CylindricalIndependent(
                r=radius, phi=angle, z=z_values, origin=(0.0, 0.0, 0.0)
            )

            my_source.angle = openmc.stats.Isotropic()
            my_source.energy = openmc.stats.Muir(
                e0=14080000.0, m_rat=5.0, kt=self.temperatures[i]
            )

            # the strength of the source (its probability) is given by
            # self.strengths
            my_source.strength = self.strengths[i]

            # append to the list of sources
            sources.append(my_source)
        return sources


def neutron_source_density(ion_density, ion_temperature):
    """Computes the neutron source density given ion density and ion
    temperature.

    Args:
        ion_density (float, ndarray): Ion density (m-3)
        ion_temperature (float, ndarray): Ion temperature (keV)

    Returns:
        float, ndarray: Neutron source density (neutron/s/m3)
    """
    return ion_density**2 * DT_xs(ion_temperature)


def DT_xs(ion_temperature):
    """Sadler–Van Belle formula
    Ref : https://doi.org/10.1016/j.fusengdes.2012.02.025

    Args:
        ion_temperature (float, ndarray): ion temperature in keV

    Returns:
        float, ndarray: the DT cross section at the given temperature
    """
    c = [
        2.5663271e-18,
        19.983026,
        2.5077133e-2,
        2.5773408e-3,
        6.1880463e-5,
        6.6024089e-2,
        8.1215505e-3,
    ]
    prod = ion_temperature * (c[2] + ion_temperature * (c[3] - c[4] * ion_temperature))
    prod *= 1 / (1.0 + ion_temperature * (c[5] + c[6] * ion_temperature))
    U = 1 - prod

    val = c[0] / (U ** (5 / 6) * ion_temperature ** (2 / 3))
    val *= np.exp(-c[1] * (U / ion_temperature) ** (1 / 3))
    return val



from openmc_plasma_source import TokamakSource
from openmc import Source
import numpy as np

import pytest
from hypothesis import given, settings, assume, strategies as st


@st.composite
def tokamak_source_strategy(draw, return_dict=False):
    """Defines a hypothesis strategy that automatically generates a TokamakSource.
    When passed `return_dict=True`, only a dict of the args is passed.
    Geometry attributes are varied, while plasma attributes are fixed.
    """
    # Used to avoid generation of inappropriate float values
    finites = {"allow_nan": False, "allow_infinity": False, "allow_subnormal": False}

    # Specify the base strategies for each geometry input
    minor_radius = draw(st.floats(min_value=0.0, max_value=100.0, **finites))
    major_radius = draw(st.floats(min_value=0.0, max_value=100.0, **finites))
    pedestal_radius = draw(st.floats(min_value=0.0, max_value=100.0, **finites))
    elongation = draw(st.floats(min_value=1.0, max_value=10.0, **finites))
    triangularity = draw(st.floats(min_value=-1.0, max_value=1.0, **finites))
    shafranov_factor = draw(st.floats(min_value=0.0, max_value=1.0, **finites))

    # Specify requirements that must be satisfied for a valid tokamak
    assume(major_radius > minor_radius)
    assume(minor_radius > pedestal_radius)
    assume(minor_radius > shafranov_factor)

    args_dict = {
        "elongation": elongation,
        "triangularity": triangularity,
        "major_radius": major_radius,
        "minor_radius": minor_radius,
        "pedestal_radius": pedestal_radius,
        "shafranov_factor": shafranov_factor,
        "ion_density_centre": 1.09e20,
        "ion_density_peaking_factor": 1,
        "ion_density_pedestal": 1.09e20,
        "ion_density_separatrix": 3e19,
        "ion_temperature_centre": 45.9,
        "ion_temperature_peaking_factor": 8.06,
        "ion_temperature_pedestal": 6.09,
        "ion_temperature_separatrix": 0.1,
        "mode": "H",
        "ion_temperature_beta": 6,
    }

    return args_dict if return_dict else TokamakSource(**args_dict)


@given(tokamak_source=tokamak_source_strategy())
@settings(max_examples=1, deadline=None)
def test_creation(tokamak_source):
    """Tests that the sources generated by TokamakSource are of type openmc.Source"""
    for source in tokamak_source.sources:
        assert isinstance(source, Source)


@given(tokamak_source=tokamak_source_strategy())
@settings(max_examples=100, deadline=None)
def test_strengths_are_normalised(tokamak_source):
    """Tests that the sum of the strengths attribute is equal to"""
    assert pytest.approx(sum(tokamak_source.strengths), 1)


@given(tokamak_source=tokamak_source_strategy())
@settings(max_examples=100, deadline=None)
def test_source_locations_are_within_correct_range(tokamak_source):
    """Tests that each source has RZ locations within the expected range.
    As the function converting (a,alpha) coordinates to (R,Z) is not bijective, we
    cannot convert back to validate each individual point. However, we can determine
    whether the generated points are contained within the shell of the last closed
    magnetic surface.
    See "Tokamak D-T neutron source models for different plasma physics confinement
    modes", C. Fausser et al., Fusion Engineering and Design, 2012 for more info.
    """
    R_0 = tokamak_source.major_radius
    A = tokamak_source.minor_radius
    El = tokamak_source.elongation
    delta = tokamak_source.triangularity

    def get_R_on_LCMS(alpha):
        """Gets R on the last closed magnetic surface for a given alpha"""
        return R_0 + A * np.cos(alpha + delta * np.sin(alpha))

    approx_lt = lambda x, y: x < y or np.isclose(x, y)
    approx_gt = lambda x, y: x > y or np.isclose(x, y)

    for source in tokamak_source.sources:
        R, Z = source.space.r.x[0], source.space.z.x[0]
        # First test that the point is contained with a simple box with
        # lower left (r_min,-z_max) and upper right (r_max,z_max)
        assert approx_gt(R, R_0 - A)
        assert approx_lt(R, R_0 + A)
        assert approx_lt(abs(Z), A * El)
        # For a given Z, we can determine the two values of alpha where
        # where a = minor_radius, and from there determine the upper and
        # lower bounds for R.
        alpha_1 = np.arcsin(abs(Z) / (El * A))
        alpha_2 = np.pi - alpha_1
        R_max, R_min = get_R_on_LCMS(alpha_1), get_R_on_LCMS(alpha_2)
        assert approx_lt(R_max, R_0 + A)
        assert approx_gt(R_min, R_0 - A)
        assert approx_lt(R, R_max)
        assert approx_gt(R, R_min)


@given(tokamak_args_dict=tokamak_source_strategy(return_dict=True))
@settings(max_examples=1, deadline=None)
def test_angles(tokamak_args_dict):
    """Checks that custom angles can be set"""
    tokamak_args_dict["angles"] = (0, 1)
    tokamak_source = TokamakSource(**tokamak_args_dict)
    assert tokamak_source.angles == (0, 1)
    for source in tokamak_source.sources:
        assert (source.space.phi.a, source.space.phi.b) == (0, 1)



from openmc_plasma_source import FusionPointSource

from openmc import Source
import pytest


def test_creation():
    my_source = FusionPointSource()
    assert isinstance(my_source, Source)


def test_wrong_fuel():
    with pytest.raises(ValueError):
        FusionPointSource(fuel="топливо")
